@@ -63,6 +63,7 @@
 #if PG_VERSION_NUM >= 160000
 #include "utils/pg_rusage.h"
 #endif
+#include "utils/timestamp.h"
 
 #include "pg_stat_kcache.h"
 
@@ -90,6 +91,11 @@ typedef uint32 pgsk_queryid;
 #define TIMEVAL_DIFF(start, end) ((double) end.tv_sec + (double) end.tv_usec / 1000000.0) \
 	- ((double) start.tv_sec + (double) start.tv_usec / 1000000.0)
 
+#if PG_VERSION_NUM < 170000
+#define MyProcNumber MyBackendId
+#define ParallelLeaderProcNumber ParallelLeaderBackendId
+#endif
+
 #if PG_VERSION_NUM < 140000
 #define ParallelLeaderBackendId ParallelMasterBackendId
 #endif
@@ -103,10 +109,12 @@ typedef enum pgskVersion
 {
 	PGSK_V2_0 = 0,
 	PGSK_V2_1,
-	PGSK_V2_2
+	PGSK_V2_2,
+	PGSK_V2_3
 } pgskVersion;
 
-static const uint32 PGSK_FILE_HEADER = 0x0d756e11;
+/* Magic number identifying the stats file format */
+static const uint32 PGSK_FILE_HEADER = 0x20240914;
 
 static struct	rusage exec_rusage_start[PGSK_MAX_NESTED_LEVEL];
 #if PG_VERSION_NUM >= 130000
@@ -135,6 +143,7 @@ typedef struct pgskEntry
 	pgskHashKey		key;		/* hash key of entry - MUST BE FIRST */
 	pgskCounters	counters[PGSK_NUMKIND];	/* statistics for this query */
 	slock_t			mutex;		/* protects the counters only */
+	TimestampTz		stats_since; /* timestamp of entry allocation */
 } pgskEntry;
 
 /*
@@ -151,13 +160,8 @@ typedef struct pgskSharedState
 
 /*---- Local variables ----*/
 
-/* Current nesting depth of ExecutorRun+ProcessUtility calls */
-static int	exec_nested_level = 0;
-
-#if PG_VERSION_NUM >= 130000
-/* Current nesting depth of planner calls */
-static int	plan_nested_level = 0;
-#endif
+/* Current nesting depth of planner/ExecutorRun/ProcessUtility calls */
+static int	nesting_level = 0;
 
 
 /* saved hook address in case of unload */
@@ -177,6 +181,10 @@ static ExecutorEnd_hook_type prev_ExecutorEnd = NULL;
 static pgskSharedState *pgsk = NULL;
 static HTAB *pgsk_hash = NULL;
 
+/*---- HOOK variables ----*/
+
+pgsk_counters_hook_type pgsk_counters_hook = NULL;
+
 /*---- GUC variables ----*/
 
 typedef enum
@@ -203,8 +211,6 @@ static bool pgsk_track_planning = false;	/* whether to track planning duration *
 	((pgsk_track == PGSK_TRACK_ALL && (level) < PGSK_MAX_NESTED_LEVEL) || \
 	(pgsk_track == PGSK_TRACK_TOP && (level) == 0))
 
-#define is_top(level) (level) == 0
-
 /*--- Functions --- */
 
 void	_PG_init(void);
@@ -213,11 +219,13 @@ extern PGDLLEXPORT Datum	pg_stat_kcache_reset(PG_FUNCTION_ARGS);
 extern PGDLLEXPORT Datum	pg_stat_kcache(PG_FUNCTION_ARGS);
 extern PGDLLEXPORT Datum	pg_stat_kcache_2_1(PG_FUNCTION_ARGS);
 extern PGDLLEXPORT Datum	pg_stat_kcache_2_2(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum	pg_stat_kcache_2_3(PG_FUNCTION_ARGS);
 
 PG_FUNCTION_INFO_V1(pg_stat_kcache_reset);
 PG_FUNCTION_INFO_V1(pg_stat_kcache);
 PG_FUNCTION_INFO_V1(pg_stat_kcache_2_1);
 PG_FUNCTION_INFO_V1(pg_stat_kcache_2_2);
+PG_FUNCTION_INFO_V1(pg_stat_kcache_2_3);
 
 static void pg_stat_kcache_internal(FunctionCallInfo fcinfo, pgskVersion
 		api_version);
@@ -254,7 +262,7 @@ static pgskEntry *pgsk_entry_alloc(pgskHashKey *key);
 static void pgsk_entry_dealloc(void);
 static void pgsk_entry_reset(void);
 static void pgsk_entry_store(pgsk_queryid queryId, pgskStoreKind kind,
-							 int level, pgskCounters counters);
+							 pgskCounters counters);
 static uint32 pgsk_hash_fn(const void *key, Size keysize);
 static int	pgsk_match_fn(const void *key1, const void *key2, Size keysize);
 
@@ -431,7 +439,7 @@ pgsk_set_queryid(pgsk_queryid queryid)
 	/* Only the leader knows the queryid. */
 	Assert(!IsParallelWorker());
 
-	pgsk->queryids[MyBackendId] = queryid;
+	pgsk->queryids[MyProcNumber] = queryid;
 }
 #endif
 
@@ -544,11 +552,13 @@ pgsk_shmem_startup(void)
 		if (fread(&temp, sizeof(pgskEntry), 1, file) != 1)
 			goto error;
 
+		/* make the hashtable entry (discards old entries if too many) */
 		entry = pgsk_entry_alloc(&temp.key);
 
 		/* copy in the actual stats */
 		entry->counters[0] = temp.counters[0];
 		entry->counters[1] = temp.counters[1];
+		entry->stats_since = temp.stats_since;
 		/* don't initialize spinlock, already done */
 	}
 
@@ -724,7 +734,7 @@ pgsk_queryids_array_size(void)
 
 static void
 pgsk_entry_store(pgsk_queryid queryId, pgskStoreKind kind,
-				 int level, pgskCounters counters)
+				 pgskCounters counters)
 {
 	volatile pgskEntry *e;
 
@@ -739,7 +749,7 @@ pgsk_entry_store(pgsk_queryid queryId, pgskStoreKind kind,
 	key.userid = GetUserId();
 	key.dbid = MyDatabaseId;
 	key.queryid = queryId;
-	key.top = is_top(level);
+	key.top = (nesting_level == 0);
 
 	/* Lookup the hash table entry with shared lock. */
 	LWLockAcquire(pgsk->lock, LW_SHARED);
@@ -813,6 +823,7 @@ static pgskEntry *pgsk_entry_alloc(pgskHashKey *key)
 		entry->counters[0].usage = USAGE_INIT;
 		/* re-initialize the mutex each time ... we assume no one using it */
 		SpinLockInit(&entry->mutex);
+		entry->stats_since = GetCurrentTimestamp();
 	}
 
 	return entry;
@@ -909,28 +920,19 @@ pgsk_planner(Query *parse,
 {
 	PlannedStmt *result;
 
-	/*
-	 * We can't process the query if no queryid has been computed.
-	 *
-	 * Note that planner_hook can be called from the planner itself, so we
-	 * have a specific nesting level for the planner.  However, utility
-	 * commands containing optimizable statements can also call the planner,
-	 * same for regular DML (for instance for underlying foreign key queries).
-	 * So testing the planner nesting level only is not enough to detect real
-	 * top level planner call.
-	 */
-	if (pgsk_enabled(plan_nested_level + exec_nested_level)
+	/* We can't process the query if no queryid has been computed. */
+	if (pgsk_enabled(nesting_level)
 		&& pgsk_track_planning
 		&& parse->queryId != UINT64CONST(0))
 	{
-		struct rusage *rusage_start = &plan_rusage_start[plan_nested_level];
+		struct rusage *rusage_start = &plan_rusage_start[nesting_level];
 		struct rusage rusage_end;
 		pgskCounters counters;
 
 		/* capture kernel usage stats in rusage_start */
 		getrusage(RUSAGE_SELF, rusage_start);
 
-		plan_nested_level++;
+		nesting_level++;
 		PG_TRY();
 		{
 			if (prev_planner_hook)
@@ -939,11 +941,11 @@ pgsk_planner(Query *parse,
 			else
 				result = standard_planner(parse, query_string, cursorOptions,
 										  boundParams);
-			plan_nested_level--;
+			nesting_level--;
 		}
 		PG_CATCH();
 		{
-			plan_nested_level--;
+			nesting_level--;
 			PG_RE_THROW();
 		}
 		PG_END_TRY();
@@ -954,22 +956,38 @@ pgsk_planner(Query *parse,
 		pgsk_compute_counters(&counters, rusage_start, &rusage_end, NULL);
 
 		/* store current number of block reads and writes */
-		pgsk_entry_store(parse->queryId, PGSK_PLAN, plan_nested_level + exec_nested_level, counters);
+		pgsk_entry_store(parse->queryId, PGSK_PLAN, counters);
 
 		if (pgsk_counters_hook)
 		    pgsk_counters_hook(&counters,
 							   query_string,
-							   plan_nested_level + exec_nested_level,
+							   nesting_level,
 							   PGSK_PLAN);
 	}
 	else
 	{
-		if (prev_planner_hook)
-			result = prev_planner_hook(parse, query_string, cursorOptions,
-									   boundParams);
-		else
-			result = standard_planner(parse, query_string, cursorOptions,
-									  boundParams);
+		/*
+		 * Even though we're not tracking planning for this statement, we
+		 * must still increment the nesting level, to ensure that functions
+		 * evaluated during planning are not seen as top-level calls.
+		 */
+		nesting_level++;
+		PG_TRY();
+		{
+			if (prev_planner_hook)
+				result = prev_planner_hook(parse, query_string, cursorOptions,
+										   boundParams);
+			else
+				result = standard_planner(parse, query_string, cursorOptions,
+										  boundParams);
+			nesting_level--;
+		}
+		PG_CATCH();
+		{
+			nesting_level--;
+			PG_RE_THROW();
+		}
+		PG_END_TRY();
 	}
 
 	return result;
@@ -979,9 +997,9 @@ pgsk_planner(Query *parse,
 static void
 pgsk_ExecutorStart (QueryDesc *queryDesc, int eflags)
 {
-	if (pgsk_enabled(exec_nested_level))
+	if (pgsk_enabled(nesting_level))
 	{
-		struct rusage *rusage_start = &exec_rusage_start[exec_nested_level];
+		struct rusage *rusage_start = &exec_rusage_start[nesting_level];
 
 		/* capture kernel usage stats in rusage_start */
 		getrusage(RUSAGE_SELF, rusage_start);
@@ -1018,7 +1036,7 @@ pgsk_ExecutorRun(QueryDesc *queryDesc,
 #endif
 )
 {
-	exec_nested_level++;
+	nesting_level++;
 	PG_TRY();
 	{
 		if (prev_ExecutorRun)
@@ -1033,11 +1051,11 @@ pgsk_ExecutorRun(QueryDesc *queryDesc,
 #else
 			standard_ExecutorRun(queryDesc, direction, count);
 #endif
-		exec_nested_level--;
+		nesting_level--;
 	}
 	PG_CATCH();
 	{
-		exec_nested_level--;
+		nesting_level--;
 		PG_RE_THROW();
 	}
 	PG_END_TRY();
@@ -1049,18 +1067,18 @@ pgsk_ExecutorRun(QueryDesc *queryDesc,
 static void
 pgsk_ExecutorFinish(QueryDesc *queryDesc)
 {
-	exec_nested_level++;
+	nesting_level++;
 	PG_TRY();
 	{
 		if (prev_ExecutorFinish)
 			prev_ExecutorFinish(queryDesc);
 		else
 			standard_ExecutorFinish(queryDesc);
-		exec_nested_level--;
+		nesting_level--;
 	}
 	PG_CATCH();
 	{
-		exec_nested_level--;
+		nesting_level--;
 		PG_RE_THROW();
 	}
 	PG_END_TRY();
@@ -1073,16 +1091,16 @@ pgsk_ExecutorEnd (QueryDesc *queryDesc)
 	struct rusage rusage_end;
 	pgskCounters counters;
 
-	if (pgsk_enabled(exec_nested_level))
+	if (pgsk_enabled(nesting_level))
 	{
-		struct rusage *rusage_start = &exec_rusage_start[exec_nested_level];
+		struct rusage *rusage_start = &exec_rusage_start[nesting_level];
 
 		/* capture kernel usage stats in rusage_end */
 		getrusage(RUSAGE_SELF, &rusage_end);
 
 #if PG_VERSION_NUM >= 90600
 		if (IsParallelWorker())
-			queryId = pgsk->queryids[ParallelLeaderBackendId];
+			queryId = pgsk->queryids[ParallelLeaderProcNumber];
 		else
 #endif
 		queryId = queryDesc->plannedstmt->queryId;
@@ -1090,12 +1108,12 @@ pgsk_ExecutorEnd (QueryDesc *queryDesc)
 		pgsk_compute_counters(&counters, rusage_start, &rusage_end, queryDesc);
 
 		/* store current number of block reads and writes */
-		pgsk_entry_store(queryId, PGSK_EXEC, exec_nested_level, counters);
+		pgsk_entry_store(queryId, PGSK_EXEC, counters);
 
 		if (pgsk_counters_hook)
 		    pgsk_counters_hook(&counters,
 							   (const char *)queryDesc->sourceText,
-							   exec_nested_level,
+							   nesting_level,
 							   PGSK_EXEC);
 	}
 
@@ -1177,6 +1195,14 @@ pg_stat_kcache_2_2(PG_FUNCTION_ARGS)
 	return (Datum) 0;
 }
 
+PGDLLEXPORT Datum
+pg_stat_kcache_2_3(PG_FUNCTION_ARGS)
+{
+	pg_stat_kcache_internal(fcinfo, PGSK_V2_3);
+
+	return (Datum) 0;
+}
+
 static void
 pg_stat_kcache_internal(FunctionCallInfo fcinfo, pgskVersion api_version)
 {
@@ -1226,12 +1252,13 @@ pg_stat_kcache_internal(FunctionCallInfo fcinfo, pgskVersion api_version)
 	{
 		Datum			values[PG_STAT_KCACHE_COLS];
 		bool			nulls[PG_STAT_KCACHE_COLS];
-		pgskCounters	tmp;
+		volatile pgskCounters   *tmp;
 		int				i = 0;
 		int				kind, min_kind = 0;
 #ifdef HAVE_GETRUSAGE
 		int64			reads, writes;
 #endif
+		TimestampTz		stats_since;
 
 		memset(values, 0, sizeof(values));
 		memset(nulls, 0, sizeof(nulls));
@@ -1246,39 +1273,40 @@ pg_stat_kcache_internal(FunctionCallInfo fcinfo, pgskVersion api_version)
 		if (api_version < PGSK_V2_2)
 			min_kind = 1;
 
+		/* copy counters to a local variable to keep locking time short */
+		{
+			volatile pgskEntry *e = (volatile pgskEntry *) entry;
+
+			SpinLockAcquire(&e->mutex);
+			tmp = e->counters;
+			stats_since = e->stats_since;
+			SpinLockRelease(&e->mutex);
+		}
+
 		for (kind = min_kind; kind < PGSK_NUMKIND; kind++)
 		{
-			/* copy counters to a local variable to keep locking time short */
-			{
-				volatile pgskEntry *e = (volatile pgskEntry *) entry;
-
-				SpinLockAcquire(&e->mutex);
-				tmp = e->counters[kind];
-				SpinLockRelease(&e->mutex);
-			}
-
 #ifdef HAVE_GETRUSAGE
-			reads = tmp.reads * RUSAGE_BLOCK_SIZE;
-			writes = tmp.writes * RUSAGE_BLOCK_SIZE;
+			reads = tmp[kind].reads * RUSAGE_BLOCK_SIZE;
+			writes = tmp[kind].writes * RUSAGE_BLOCK_SIZE;
 			values[i++] = Int64GetDatumFast(reads);
 			values[i++] = Int64GetDatumFast(writes);
 #else
 			nulls[i++] = true; /* reads */
 			nulls[i++] = true; /* writes */
 #endif
-			values[i++] = Float8GetDatumFast(tmp.utime);
-			values[i++] = Float8GetDatumFast(tmp.stime);
+			values[i++] = Float8GetDatumFast(tmp[kind].utime);
+			values[i++] = Float8GetDatumFast(tmp[kind].stime);
 			if (api_version >= PGSK_V2_1)
 			{
 #ifdef HAVE_GETRUSAGE
-				values[i++] = Int64GetDatumFast(tmp.minflts);
-				values[i++] = Int64GetDatumFast(tmp.majflts);
-				values[i++] = Int64GetDatumFast(tmp.nswaps);
-				values[i++] = Int64GetDatumFast(tmp.msgsnds);
-				values[i++] = Int64GetDatumFast(tmp.msgrcvs);
-				values[i++] = Int64GetDatumFast(tmp.nsignals);
-				values[i++] = Int64GetDatumFast(tmp.nvcsws);
-				values[i++] = Int64GetDatumFast(tmp.nivcsws);
+				values[i++] = Int64GetDatumFast(tmp[kind].minflts);
+				values[i++] = Int64GetDatumFast(tmp[kind].majflts);
+				values[i++] = Int64GetDatumFast(tmp[kind].nswaps);
+				values[i++] = Int64GetDatumFast(tmp[kind].msgsnds);
+				values[i++] = Int64GetDatumFast(tmp[kind].msgrcvs);
+				values[i++] = Int64GetDatumFast(tmp[kind].nsignals);
+				values[i++] = Int64GetDatumFast(tmp[kind].nvcsws);
+				values[i++] = Int64GetDatumFast(tmp[kind].nivcsws);
 #else
 				nulls[i++] = true; /* minflts */
 				nulls[i++] = true; /* majflts */
@@ -1291,10 +1319,13 @@ pg_stat_kcache_internal(FunctionCallInfo fcinfo, pgskVersion api_version)
 #endif
 			}
 		}
+		if (api_version >= PGSK_V2_3)
+			values[i++] = TimestampTzGetDatum(stats_since);
 
 		Assert(i == (api_version == PGSK_V2_0 ? PG_STAT_KCACHE_COLS_V2_0 :
 					 api_version == PGSK_V2_1 ? PG_STAT_KCACHE_COLS_V2_1 :
 					 api_version == PGSK_V2_2 ? PG_STAT_KCACHE_COLS_V2_2 :
+					 api_version == PGSK_V2_3 ? PG_STAT_KCACHE_COLS_V2_3 :
 					 -1 /* fail if you forget to update this assert */ ));
 
 		tuplestore_putvalues(tupstore, tupdesc, values, nulls);
